"""
TABI - UCSC iGEM 2023
Modified from ORFfinder Class written for Winter 2022 BME160
BME160 Professor: David L. Bernick
"""

# ORF finder used to parse FastA format genome files.
# Optional parameters:
#     longestGene -> only reports longest genes in an ORF
#     minGene -> only reports genes of this size or larger
#     starts -> define set of valid start codons
#     stop -> define set fo valid stop codons

class ORFfinder:
    """
    ORF finder class that scans through a genome and reports ORFs from all frames
    
    Data is reported in a list of lists in 
    
    format: list[[startPos,stopPos,length,frame]]
    """

    def __init__(self,seq: str,longestGene=False,minGene = 100,starts: set = {"ATG"},stops: set = {"TAA","TAG","TGA"}):
        """
        Usage:
        gene_finder = ORFfinder(seq,**kwargs)
        
        gene_candidates = gene_finder.get_genes()
        """
        assert minGene >= 0
        self.geneCandidates = []
        format = seq.strip().upper().replace(" ",'') 
        flipTable = format[::-1].maketrans("ATCG", "TAGC") # Builds reverse compliment strand
        
        self.fiveThree = format 
        self.threeFive = format[::-1].translate(flipTable) 
        self.validStart = starts
        self.validStops = stops
        self.minGene = minGene
        self.longestGene = longestGene
        
        self._geneFinder()

    def _geneFinder(self,seq,rev,longGene):
        """
        Helper function that collects starts and identifies stops
        """
        start = set()
        for frame in range(0,3):
            start.clear()
            for i in range(frame,len(seq),3):
                codon = seq[i:i+3]

                if codon in self.validStart:
                    start.add(i)

                if codon in self.validStops and len(start) > 0:
                    _frame = -1-frame if rev else frame+1 
                    self._addGene(start,i,_frame,longGene)
                    start = set() 
                    

            if len(start) > 0: 
                _frame = -1-frame if rev else frame+1
                self._addGene(start,len(seq)-3,_frame,longGene)


    def _addGene(self,start,i,frame,longGene):
        """
        Helper Function
        Does all gene-candidate calculations using [start] [stop index] [frame #] [longest gene flag]
        """
        for s in sorted(start):
            length = i - s + 3
            if frame < 0: # Handles reverse strand calcs
                startPos = len(self.fiveThree) - i -2
                stopPos = len(self.fiveThree) - s
            else: # Handles top strand
                startPos = s + 1
                stopPos = i + 3
            if length >= self.minGene:
                self.geneCandidates.append([startPos, stopPos, length, frame])
            if longGene: # LongestGene flag
                return


    def _geneFinder(self):
        """
        Main geneFinder function. Runs hidden _geneFinder() function to parse through both top/bottom strand
        """
        self._geneFinder(self.fiveThree,False,self.longestGene)
        self._geneFinder(self.threeFive,True,self.longestGene)
        
        #sorts geneCandidates list, sorting by gene length
        self.formatted = [x for x in sorted(self.geneCandidates, key=lambda entry: entry[2], reverse=True)] 
        return self.formatted

    def get_genes(self):
        return self.geneCandidates


"""
TABI - UCSC iGEM 2023

StealthGenome: Prep Genome for Stealth and Codon Frequency analysis

PlasmidParser: Identifies Stealth-mutable regions of an annotated plasmid

Author: Tyler Gaw
"""
# Usage to be written, don't wanna rn

from Bio import SeqIO as _SeqIO
from Bio.SeqFeature import SimpleLocation as _SimpleLocation
from Bio.Seq import Seq as _Seq
from Bio.SeqRecord import SeqRecord as _SeqRecord
from typing import Iterator
from chameleontools.ORFfinder import ORFfinder as _ORFfinder
from chameleontools.FastAreader import FastAreader as _FastAreader
import sys as _sys

class StealthGenome:
    """
    Reads genome file in genbank or FASTA format,
    
    Stores Genome sequence and CDS regions
    """

    def _validArg(self,genome: str):
        'helper func: checks valid args'
        g_valid = False,False,False
        g_valid= genome.endswith('.fasta') or genome.endswith('.gb') or genome.endswith('.gbk')
        return g_valid

    def _fastaGenome(self,fasta: str):
        reader = _FastAreader(fasta)
        CDS_records = []
        sequence = []
        for frag,(header,seq) in enumerate(reader.readFasta()):
            genome = _Seq(seq)
            orfs = _ORFfinder(seq,longestGene=True,minGene=200).geneCandidates #array of estimated genes
            for i,j in enumerate(orfs):  
                'Converting ORFfinder output to Biopython SimpleLocation'
                loc = _SimpleLocation(j[0]-1, j[1], strand = 1 if j[3] > 0 else -1)
                
                'Building Biopython SeqRecord for CDS'
                rec = _SeqRecord(loc.extract(genome))
                rec.description = "Predicted Protein Coding Sequence"
                rec.name = 'Predicted CDS HEADER'
                rec.id = f'FRAG_{frag} CDS_{i} for CODON STATS'
                
                CDS_records.append(rec)
            sequence.append(genome)
        return CDS_records,sequence    
            
    def _get_SEQ_name(self,infile: str) -> str:
        if infile != None:
            if infile.endswith(".gb") or infile.endswith(".gbk"):
                self.outfile = infile[:-3] if infile.endswith(".gb") else infile[:-4]
                return self.outfile+'_SEQ.fasta'
            else:
                self.outfile = infile[:-6]
                return infile

    def _get_CDS_name(self) -> str:
        return self.outfile+'_CDS.fasta'
  
############################################################################################
# ^^^ Private Helper Functions ^^^
############################################################################################

    def __init__(self, genome_infile) -> None:
        '''
        Constructor: Validates inputs, writes to files in FASTA format, saves file names
        
        Can be used for Stealth_CDS_Extract to produce a CDS generator for input file

        Args: 
        genome_infile (str): genome file in genbank or FASTA format

        Class Attr.:
        
        self.genome_sequence -> Biopython _Seq() of input genome | array of _Seq() if multiple fragment FastA file
        
        self.cds_sequence -> Iterator of Biopython SeqRecord for each CDS | array of SeqRecord if using FastA file
        '''
        if genome_infile != None:
            
            g_bool = self._validArg(genome_infile)
            
            if not g_bool:
                print(f"Genome File needs to be FASTA file (.fasta) or GenBank file (.gb || .gbk). Got {genome_infile}",file = sys.stderr)
                exit()

            if genome_infile.endswith(".fasta"):
                self.genome_sequence,self.cds_sequence =  self._fastaGenome(genome_infile)
                
            else:
                self.cds_sequence = self.Stealth_CDS_Extract(genome_infile)
        
    def Stealth_CDS_Extract(self,gbfile: str) -> Iterator[SeqRecord]:
        """
        Reads a genbank file and yields CDS/ORF features for Codon Optimization

        A generator that yields CDS features in FASTA format

        Used to produce data for codon usage statistics
        """
        gb = __SeqIO.parse(gbfile,'genbank') 
        count = 0 
        for rec in gb:
            self.genome_sequence = rec.seq #define self.genome_sequence from genbank record
            for feature in rec.features:
                if feature.type == 'CDS' or feature.type == 'ORF':
                    count += 1
                    out = feature.extract(rec)

                    '''Set the description to 'product' qualifier if avaiable. Else use 'label' qualifier or [null] if labels are not present'''
                    out.description = feature.qualifiers.get('product',feature.qualifiers.get("label",["null"]))[0]

                    out.name = 'TEMP CDS HEADER'
                    out.id = f'CDS_{count} CODON STATS'
                    yield out
            return #Only accepts single entry genbank records

class PlasmidParser:
    """
    Reads in an annotated plasmid file in genbank format,
    parses mutable regions from CDS/ORF annotations,
    takes statistics for total mutable range
    """

    def _validArg(self, plasmid_infile: str) -> bool:
        'helper func: checks for valid input'
        return plasmid_infile.endswith(".gb") or plasmid_infile.endswith('.gbk')
   
    def _parsePlasmid(self, gbfile: str) -> tuple[list[SimpleLocation],list[SimpleLocation],Seq]:
        'helper func: Filters for CDS and noncoding sequence'
        avoid,cds = [],[]
        gb = _SeqIO.parse(gbfile,'genbank') 
        for rec in gb:
            sequence = rec.seq
            for feature in rec.features:
                if feature.type in {'source','gene'}:
                    continue
                if (feature.type in {'CDS','ORF'}):
                    cds.append((feature.location, feature.location.start % 3))
                else:
                    avoid.append((feature.location, feature.location.start % 3))
            self.sequence = rec.seq
            break #Only handles single entry genbank records
        return avoid,cds,sequence

    def _trimStart(self, loc_list: list[SimpleLocation], cds: SimpleLocation)  -> list[SimpleLocation]:
        'helper func: trims the start codon of CDS, handles fwd/rev strands'
        ref = _SimpleLocation(cds.start,cds.start+3,cds.strand) if cds.strand > 0 else _SimpleLocation(cds.end-3,cds.end,cds.strand)
        for i in range(len(loc_list)):
            loc = loc_list[i]
            if loc.start < ref.start and loc.end > ref.end:
                loc_list[i] = _SimpleLocation(loc.start,ref.start,loc.strand)
                loc_list.insert(i+1,_SimpleLocation(ref.end,loc.end,loc.strand)) 
                break  
            if loc.start in ref:
                loc_list[i]= _SimpleLocation(ref.end,loc.end,loc.strand)
                break
            elif loc.end in ref or loc.end == ref.end:
                loc_list[i] = _SimpleLocation(loc.start,ref.start,loc.strand)
                break
        return

    def _frameAdj(self, loc_list: list[SimpleLocation], frame: int) -> SimpleLocation:
        'helper func: trims mutable regions to lie in frame'
        new = []
        for loc in loc_list:
            st = loc.start
            ed = loc.end
            for _ in range(1,3):
                if st % 3 != frame :
                    st += 1
                if ed % 3 != frame:
                    ed -= 1
            if (ed-st) < 3:
                continue
            adj = _SimpleLocation(st,ed,loc.strand)
            new.append(adj)
        return new

    def _removeOverlap(self, loc_list: list[SimpleLocation],cds: SimpleLocation) -> list[SimpleLocation]:
        'helper func: removes overlapping cds regions '
        ret = []
        for loc in loc_list:
            new = [[loc.start,loc.end]]
            if loc.start in cds and (loc.end in cds or loc.end == cds.end):
                'completely removes region'
                continue
            if loc.start < cds.start and loc.end > cds.end:
                'splits range into two'
                dummy = [cds.end,loc.end]
                new[-1][1] = cds.start
                new.append(dummy) 
            elif loc.start in cds:
                'truncates start'
                new[-1][0] = cds.end   
            elif loc.end in cds or loc.end == cds.end:
                'truncates end'
                new[-1][1] = cds.start+1
            ret.extend([_SimpleLocation(i,j,loc.strand) for i,j in new])
        return ret

    def _windowCorrection(self, old: list[list[SimpleLocation],int,SimpleLocation], st_bound: int, ed_bound: int) -> list[list[SimpleLocation],int,SimpleLocation]:
        'helper func: corrects for out-frame overlapping CDS'
        loc,fm,parent = old
        new = _SimpleLocation(st_bound,ed_bound,parent.strand)
        dummy = []
        
        for i in loc:
            if st_bound > i.end:
                continue
            if i.start in new and (i.end in new or i.end == new.end):
                dummy.append(i)
            elif i.start in new:
                dummy.append(_SimpleLocation(i.start,ed_bound,i.strand))     
                continue
            elif i.end in new or i.end == new.end:
                dummy.append(_SimpleLocation(st_bound,i.end,i.strand))
                continue 

            
        return [dummy,fm,new]

############################################################################################
# Private Helper Functions
############################################################################################

    def __init__(self, plasmid_infile: str) -> None:
        '''
        Constructor: Validates inputs, parses mutable regions

        Args: 
        plasmid_infiled (str): genome file in FASTA or genbank format
        
        Class Attr.:
        
        self.plasmid_sequence -> Biopython _Seq() of input Plasmid

        self.mutable_regions -> array of Biopython _SimpleLocation() ranges of mutable codons
        
        self.total_mutable -> final mutable regions on plasmid
        
        self.removed -> number of removed basepairs from total CDS coverage
        
        self.total_coverage -> Total CDS coverage before extracting mutable regions
        '''
        p_bool = self._validArg(plasmid_infile)
        if not p_bool:
            print(f"Plasmid Input File needs to be GenBank file (.gb || .gbk). Got: {plasmid_infile}",file=sys.stderr)
            exit()
        
        'Did this cause it was ugly when using tuple unpacking'
        temp_parse = self._parsePlasmid(plasmid_infile)
        
        temp = self.defineMutable(temp_parse[0:2])
        
        self.plasmid_sequence = temp_parse[2]
        
        self.mutable_regions = temp[0]
        self.total_mutable = temp[1]
        self.removed = temp[2]
        
        self.total_coverage = self.total_mutable+self.removed
        
    def defineMutable(self,location: tuple[list[SimpleLocation]]) -> list[SimpleLocation]:
        'Finds non-overlapped mutable regions in an annotated plasmid'

        avoid,cdsRegions = location
        temp = [] #[[[SimpLocation List],frame,parent CDS]]

        'Removes mutable cds regions from overlapping non-coding sequence'
        for cds,frame in cdsRegions:
            if len(cds) <= 3:
                continue
            bounds = [[cds.start,cds.end]]
            for loc,_ in avoid:
                if cds.start > loc.start and cds.end < loc.end:
                    'case: interior ORF to non-coding region'
                    bounds = [[-1,-1]]
                    break
                if bounds[-1][0] < loc.start and bounds[-1][1] >loc.end:
                    'case: non-coding region interior to ORF'
                    new_bound = [loc.end,cds.end]
                    bounds[-1][1] = loc.start
                    bounds.append(new_bound)
                elif bounds[-1][0] in loc:
                    'case: clipped start'
                    bounds[-1][0] = loc.end
                elif bounds[-1][1] in loc or bounds[-1][1] == loc.end:
                    'case: clipped end'
                    bounds[-1][1] = loc.start
            if (bounds[-1][1] - bounds[-1][0]) <= 3:
                del bounds[-1]
            if len(bounds) <= 0:
                continue
            mut_range = [_SimpleLocation(start=rng[0],end=rng[1],strand=cds.strand) for rng in bounds ]
            temp.append([mut_range,frame,cds])
        
        'Trimming start codons'   
        for i in range(len(temp)):
            tLoc = temp[i][0]
            frm = temp[i][1]
            reg = temp[i][2]
            window = temp[i+1:]
            for j in range(len(window)):
                _,parent_frame,del_cds = window[j]
                if (del_cds.start in reg and del_cds.strand > 0) or (del_cds.end in reg and del_cds.strand < 0):
                    self._trimStart(temp[i][0],del_cds)
                    tLoc = temp[i][0]
            if reg.strand > 0:
                trimmed = _SimpleLocation(reg.start+3,reg.end,reg.strand)
                if tLoc[0].start == reg.start:
                    temp[i][0][0] = _SimpleLocation(trimmed.start,tLoc[0].end,tLoc[0].strand)
            else:
                trimmed = _SimpleLocation(reg.start,reg.end-3,reg.strand)
                if tLoc[-1].end == reg.end:
                    temp[i][0][-1] = _SimpleLocation(tLoc[-1].start,trimmed.end,tLoc[-1].strand)
        
        'sorts regions by CDS length, compares larger CDS to smaller CDS to remove overlap'
        temp = sorted(temp,key = lambda x: len(x[2]), reverse=True)
        
        'remove overlaps'
        for i in range(len(temp)):
            frag = temp[i][0]
            frm = temp[i][1]
            region = temp[i][2]
            window = temp[i+1:]
            if region == None:
                continue
            for j in range(len(window)):
                _,parent_frame,parent_cds = window[j]
                if parent_cds == None:
                    continue
                if parent_cds.start in region or parent_cds.end in region:
                    if parent_frame != frm:
                        temp[i][0] = self._removeOverlap(frag,parent_cds)
                        frag = temp[i][0]
                if parent_cds.start in region and parent_cds.end in region:
                    temp[i+j+1][2] = None
                elif parent_cds.start in region:
                    temp[i+j+1] = self._windowCorrection(window[j],region.end,parent_cds.end)
                elif parent_cds.end in region:
                    temp[i+j+1] = self._windowCorrection(window[j],parent_cds.start,region.start)
        
        'frame adjusts mutable fragments to lie in frame of parent CDS'
        for i in range(len(temp)):
            temp[i][0] = self._frameAdj(temp[i][0],temp[i][1])
        
        mutable_regions = sorted([region for CDS,_,tf in temp for region in CDS if tf is not None],key = lambda x: x.start)
        
        'mutable region statistics - total covered region and mutable regions'
        total_coverage = sum([len(x[-1]) for x in temp if x[-1] is not None])
        removed = self.total_coverage - sum([len(x) for x in mutable_regions])
        return mutable_regions,total_coverage,removed




#!/usr/bin/env python3
# Name: Allison Jaballas (acjaball)

##########################################################################
############################## FASTA READER ##############################
##########################################################################

import sys
class FastAreader:
    ''' 
    Define objects to read FastA files.
    
    instantiation: 
    thisReader = FastAreader ('testTiny.fa')
    usage:
    for head, seq in thisReader.readFasta():
        print (head,seq)
    '''
    def __init__ (self, fname=None):
        '''contructor: saves attribute fname '''
        self.fname = fname
            
    def doOpen (self):
        ''' Handle file opens, allowing STDIN.'''
        if self.fname is None:
            return sys.stdin
        else:
            return open(self.fname)
        
    def readFasta (self):
        ''' Read an entire FastA record and return the sequence header/sequence'''
        header = ''
        sequence = ''
        
        with self.doOpen() as fileH:
            
            header = ''
            sequence = ''
            
            # skip to first fasta header
            line = fileH.readline()
            while not line.startswith('>') :
                line = fileH.readline()
            header = line[1:].rstrip()

            for line in fileH:
                if line.startswith ('>'):
                    yield header,sequence
                    header = line[1:].rstrip()
                    sequence = ''
                else :
                    sequence += ''.join(line.rstrip().split()).upper()

        yield header,sequence

########################################################################
############################## NUC PARAMS ##############################
########################################################################

class NucParams:
    '''
    Define objects to store counts of codons and amino acids from input sequence.
    '''
    
    dnaCodonTable = {
    # Second Base
    # T             C             A             G
    #T
    'TTT': 'Phe', 'TCT': 'Ser', 'TAT': 'Tyr', 'TGT': 'Cys',
    'TTC': 'Phe', 'TCC': 'Ser', 'TAC': 'Tyr', 'TGC': 'Cys',
    'TTA': 'Leu', 'TCA': 'Ser', 'TAA': '---', 'TGA': '---',
    'TTG': 'Leu', 'TCG': 'Ser', 'TAG': '---', 'TGG': 'Trp',
    #C 
    'CTT': 'Leu', 'CCT': 'Pro', 'CAT': 'His', 'CGT': 'Arg',
    'CTC': 'Leu', 'CCC': 'Pro', 'CAC': 'His', 'CGC': 'Arg',
    'CTA': 'Leu', 'CCA': 'Pro', 'CAA': 'Gln', 'CGA': 'Arg',
    'CTG': 'Leu', 'CCG': 'Pro', 'CAG': 'Gln', 'CGG': 'Arg',
    #A
    'ATT': 'Ile', 'ACT': 'Thr', 'AAT': 'Asn', 'AGT': 'Ser',
    'ATC': 'Ile', 'ACC': 'Thr', 'AAC': 'Asn', 'AGC': 'Ser',
    'ATA': 'Ile', 'ACA': 'Thr', 'AAA': 'Lys', 'AGA': 'Arg',
    'ATG': 'Met', 'ACG': 'Thr', 'AAG': 'Lys', 'AGG': 'Arg',
    #G
    'GTT': 'Val', 'GCT': 'Ala', 'GAT': 'Asp', 'GGT': 'Gly',
    'GTC': 'Val', 'GCC': 'Ala', 'GAC': 'Asp', 'GGC': 'Gly',
    'GTA': 'Val', 'GCA': 'Ala', 'GAA': 'Glu', 'GGA': 'Gly',
    'GTG': 'Val', 'GCG': 'Ala', 'GAG': 'Glu', 'GGG': 'Gly'
    }

    def __init__ (self, inString=''):
        ''' Set up codon composition, amino acid composition, and nucleotide composition dictionaries for future use.'''
        self.inString = inString.upper()
        
        self.codonComp = {codon: 0 for codon in self.dnaCodonTable.keys()} #set up empty (0) dict for codon counts

        self.aaComp = {
            'Phe': 0, 'Leu': 0, 'Ser': 0, 'Tyr': 0, '---': 0, 'Cys': 0, 'Trp': 0, 'Pro': 0,
            'His': 0, 'Gln': 0, 'Arg': 0, 'Ile': 0, 'Met': 0, 'Thr': 0, 'Asn': 0, 'Lys': 0,
            'Val': 0, 'Ala': 0, 'Asp': 0, 'Glu': 0, 'Gly': 0
            }
        
        self.nucComp = { 'A': 0, 'C': 0, 'G' : 0, 'T' : 0, 'U' : 0, 'N' : 0 } 

        self.addSequence(inString)
        
    def addSequence (self, inSeq):
        '''Add nucleotide count totals and decode codon information to appropriate dictionaries.'''
        inSeq = self.inString.join(inSeq.split()).upper()

        # count totals of each nucleotide from file input
        for nuc in self.nucComp.keys():
            self.nucComp[nuc] += inSeq.count(nuc)

        # to decode codon -> AA
        AA = [ inSeq[n:n+3] for n in range(0, len(inSeq), 3) ]
        
        # add to codonComp + aaComp dictionaries!
        for codon in AA:
            if codon in self.codonComp: 
                self.codonComp[codon] += 1
                aa = self.dnaCodonTable[codon]
                self.aaComp[aa] += 1
       
    def nucComposition(self):
        '''Return dictionary counts of valid nucleotides from file input.'''
        return self.nucComp
    def codonComposition(self):
        '''Return dictionary counts of codons from file input.'''
        return self.codonComp
    def nucCount(self):
        '''Return sum of all the valid nucleotides based on nucComposition.'''
        return sum(self.nucComp.values())

#!/usr/bin/env python3
# Name: Allison Jaballas (acjaball)

#############################################################################
############################## GENOME ANALYZER ##############################
#############################################################################

class genomeAnalyzer:
    '''
    GenomeAnalyzer calls the methods from the classes imported from sequenceAnalyzer to calculate the
    file input's sequence length, GC content, and relative codon usages. This program prepares the
    summaries and final display of all the data.

    Dictionaries are as follows:

    totalCodonsPerAA = {codon:#count}
    codonFractionUsed = {codon:fraction}

    '''
    def __init__ (self, filename = None):
        '''Constructor: saves attribute filename.'''
        self.calculations(filename)
        
    def calculations (self,fastafile):
        '''Calculate the file input's sequence length and relative codon usages.'''
        myReader = FastAreader(fastafile)
        myNuc = NucParams()
        for head,seq in myReader.readFasta():
            myNuc.addSequence(seq)
            
        seqLength = myNuc.nucCount()/1000000
        #print('sequence length = {:.2f} Mb'.format(seqLength))
        #print()

        codonCount = myNuc.codonComposition() #shorthand version
        
        # calculate fraction used
        # set up dictionaries
        # totalCodonsPerAA gets the total # of codons for each aa
        codonFractionUsed = {}
        totalCodonsPerAA = {aa: sum(count for codon, count in myNuc.codonComp.items() if myNuc.dnaCodonTable[codon] == aa)for aa in myNuc.aaComp}
        
        codonTot = 0

        for codon, count in myNuc.codonComp.items(): # calculate fraction
            aa = myNuc.dnaCodonTable[codon] #the aa of each codon
                
            if totalCodonsPerAA[aa] > 0:
                codonFractionUsed[codon] = count/totalCodonsPerAA[aa] # builds the dict codonFractionUsed

            codonTot += count
                                    
        for codons in sorted(codonCount.keys()): # pull together all information and prepare to print
            codCount = codonCount[codons]
            fracCount = codonFractionUsed[codons]
            aa = myNuc.dnaCodonTable[codons] if myNuc.dnaCodonTable[codons] != "---" else "End"
            aaCount = myNuc.aaComp[aa] if aa != 'End' else myNuc.aaComp['---']

            # codon freq calc: #codons/totalAA
            codonUsage = (codCount/codonTot) * 1000

            # print everything together
            print('{:5} {:5} {:5d} {:5.2f} {:5.2f}'.format(aa, codons, codCount, codonUsage, fracCount))
   
genomeAnalyzer("microcystis_blast_contigs_CDS.fasta")



"""
TABI - UCSC iGEM 2023

Classes to parse Stealth output and isolate restricted motifs

Author: Tyler Gaw
"""

# Takes in a file path to Stealth output and generates a set of expanded degenerate motifs
# If no file_path is specified, reads from STDIN
# Usage Example:
#     motif_set_1 = ParseStealth(file_path_1)
#     motif_set_2 = PalindromeParseStealth(file_path_2)
#     conserved_motifs = motif_set_1.intersect(motif_set_2) -> set

from typing import Iterator
import sys

class ParseStealth(set):
    '''
    Takes stealth output data as a text file and processes a set of motifs.
    '''
    degenerateBases = { # Degenerate base pair dictionary
            'R': ['A', 'G'],
            'Y': ['C', 'T'],
            'M': ['A', 'C'],
            'K': ['G', 'T'],
            'S': ['C', 'G'],
            'W': ['A', 'T'],
            'H': ['A', 'C', 'T'],
            'B': ['C', 'G', 'T'],
            'V': ['A', 'C', 'G'],
            'D': ['A', 'G', 'T'],
            'N': ['A', 'C', 'G', 'T']
            }

    def __init__(self, file_path: str):
        '''
        Constructor: Inherits from set\n
        args : <file_path> - File path to stealth output\n
        Output: Set containing expanded motifs
        '''
        super().__init__()
        self._readIn(file_path)
    
    def _readIn(self,file_path: str) -> set:
        '''
        helper func - from file reads all motifs into set
        '''
        file_p = open(file_path,'r') if type(file_path) == str else sys.stdin
        with file_p as fd:
            while line := fd.readline():
                if not line.startswith("N ="):
                    continue
                line = fd.readline().strip().split()
                for expanded_motif in self._permute(line[0]):
                    self.add(expanded_motif)
    
    def _permute(self,string, cur_idx=0, cur_perm="") -> Iterator[str]:
        '''
        helper func - generates permutations of all degenerate sequences
        '''
        if cur_idx == len(string):
            yield cur_perm
            if cur_perm == string:
                return
        else:
            cur_let = string[cur_idx]
            if cur_let in self.degenerateBases:
                subs = self.degenerateBases[cur_let]
                for sub in subs:
                    yield from self._permute(string,cur_idx + 1,cur_perm + sub)
            else:
                yield from self._permute(string,cur_idx + 1,cur_perm + cur_let)

class PalindromeParseStealth(ParseStealth):
    
    def __init__(self, file_path: str) -> set:
        '''
        Constructor: Inherits from ParseStealth --> Set\n
        args : <file_path> - File path to stealth output\n
        Output: Set containing expanded RC palindrome motifs
        '''
        super().__init__(file_path)
        
    def _readIn(self,file_path: str) -> set: 
        '''
        helper func overload - from file reads palindromes into set
        '''
        file_p = open(file_path,'r') if type(file_path) == str else sys.stdin
        with file_p as fd:
            
            while line := fd.readline():
                
                if not line.startswith("N ="): # Skip to first motif
                    continue
                
                line = fd.readline().strip().split()
                if line[-1] == 'Palindrome':
                    for expanded_motif in self._permute(line[0]):
                        if self._revComp(expanded_motif):
                            self.add(expanded_motif)

    def _revComp(self,seq: str) -> bool:
        '''
        helper function - boolean check on reverse compliment
        '''
        t = {'A':'T','C':'G','T':'A','G':'C'}
        if len(seq) % 2 == 0:
            return ''.join([t[i] for i in seq[::-1]]) == seq
        else:
            return "".join([t[i] for i in seq[:len(seq)//2:-1]]) == seq[:len(seq)//2]
    


#!/usr/bin/env python3
"""
TABI - UCSC iGEM 2023

Commandline interface for parsing Stealth output

Standalone program for ParseStealth.py

Author: Tyler Gaw
"""

# Usage: ./pipeline.py -i <input file | default= stdin> -o <outfile | default stdout> -[optional]
#         -s [sorted | default= False]
#         -p [RC paindromes only | default= False]
#         -d DNAworks compatible output


import argparse,sys
from chameleontools.ParseStealth import ParseStealth,PalindromeParseStealth

def writeFile(motif: set, outfile: str, sortBool: bool, header: str, dnaWorks: bool):
    '''
    writes stealth output to DNAworks format file
    '''
    file_p = open(outfile,"w") if type(outfile) == str else sys.stdout
    with file_p as fd:
        fd.write(f"Run Options: {header}\n")
        if sortBool:    
            'explicit sort by length first, then alphabetical'
            for i,seq in enumerate(sorted(list(motif),key= lambda x : (len(x),x))):
                descriptor = f" re{i}" if dnaWorks else f" [temp]"
                fd.write(f"{descriptor}\t{seq}\n")
        else:
            for seq in list(motif):
                descriptor = f" re{i}" if dnaWorks else f" [temp]"
                fd.write(f"{descriptor}\t{seq}\n")
    

def main():
    '''
    Commandline parse, executes file write
    '''
    parser = argparse.ArgumentParser(description= "Reads in Stealth file, outputs motifs in formatted file",usage= f"{sys.argv[0]} -i <input file | default= stdin> -o <outfile | default stdout> -[optional]\n\t-s [sorted | default= False]\n\t-p [RC paindromes only | default= False]\n\t-d DNAworks compatible output")
    parser.add_argument("--infile",'-i',default=None,type=str,action='store',help='input file directory')
    parser.add_argument("--outfile",'-o',default=None,type=str,action='store',help='output file')
    parser.add_argument("--sorted",'-s',default=False,action='store_true',help='sort output')
    parser.add_argument("--palindrome",'-p',default=False,action='store_true',help='palindrome output')
    parser.add_argument("--dnaWorks",'-d',default=False,action='store_true',help='DNAWorks compatible output')
    args = parser.parse_args()
    conserved = PalindromeParseStealth(args.infile) if args.palindrome else ParseStealth(args.infile)
    writeFile(conserved,args.outfile,args.sorted," ".join(sys.argv),args.dnaWorks)


if __name__ == "__main__":
    main()










# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['chameleontools']

package_data = \
{'': ['*']}

install_requires = \
['biopython>=1.81,<2.0', 'scipy>=1.10.1,<2.0.0']

entry_points = \
{'console_scripts': ['pstealth = chameleontools.pipeline:main']}

setup_kwargs = {
    'name': 'chameleontools',
    'version': '0.0.1',
    'description': 'A Stealth-based pipeline that optimizes plasmids for bacterial transformations in non-model organisms.',
    'long_description': "# Team UCSC 2023 Software Tool\n\nIf you team competes in the [**Software & AI** village](https://competition.igem.org/participation/villages) or wants to\napply for the [**Best Software Tool** prize](https://competition.igem.org/judging/awards), you **MUST** host all the\ncode of your team's software tool in this repository, `main` branch. By the **Wiki Freeze**, a\n[release](https://docs.gitlab.com/ee/user/project/releases/) will be automatically created as the judging artifact of\nthis software tool. You will be able to keep working on your software after the Grand Jamboree.\n\n> If your team does not have any software tool, you can totally ignore this repository. If left unchanged, this\nrepository will be automatically deleted by the end of the season.\n\n\n\n## Description\nLet people know what your project can do specifically. Provide context and add a link to any reference visitors might\nbe unfamiliar with (for example your team wiki). A list of Features or a Background subsection can also be added here.\nIf there are alternatives to your project, this is a good place to list differentiating factors.\n\n## Installation\nWithin a particular ecosystem, there may be a common way of installing things, such as using Yarn, NuGet, or Homebrew.\nHowever, consider the possibility that whoever is reading your README is a novice and would like more guidance. Listing\nspecific steps helps remove ambiguity and gets people to using your project as quickly as possible. If it only runs in a\nspecific context like a particular programming language version or operating system or has dependencies that have to be\ninstalled manually, also add a Requirements subsection.\n\n## Usage\nUse examples liberally, and show the expected output if you can. It's helpful to have inline the smallest example of\nusage that you can demonstrate, while providing links to more sophisticated examples if they are too long to reasonably\ninclude in the README.\n\n## Contributing\nState if you are open to contributions and what your requirements are for accepting them.\n\nFor people who want to make changes to your project, it's helpful to have some documentation on how to get started.\nPerhaps there is a script that they should run or some environment variables that they need to set. Make these steps\nexplicit. These instructions could also be useful to your future self.\n\nYou can also document commands to lint the code or run tests. These steps help to ensure high code quality and reduce\nthe likelihood that the changes inadvertently break something. Having instructions for running tests is especially\nhelpful if it requires external setup, such as starting a Selenium server for testing in a browser.\n\n## Authors and acknowledgment\nShow your appreciation to those who have contributed to the project.\n",
    'author': 'Your Name',
    'author_email': 'you@example.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.igem.org/2023/software-tools/ucsc',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import sys
import chameleontools.StealthGenome as _GenomeFile

class _pipeline:
    
    def _parseInputGenome(self, infile: str) -> tuple[str,str]:
        pass
    
    def __init__(self) -> None:
        pass
    
    
        